# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from ipywidgets import Layout

default_layout = Layout(width='100%')
default_style = {'description_width': 'auto'}



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""Tools for handling statistical uncertainties.

This module provides tools for handling statistical uncertainties in the context of
data reduction. Handling variances during broadcast operations is not handled
correctly by Scipp because correlations are not tracked.
See https://doi.org/10.3233/JNR-220049 for context.

This module provides three ways of handling variances during broadcast operations,
defined by :py:class:`UncertaintyBroadcastMode`.
The recommended use of this module is via the :py:func:`broadcast_uncertainties`
helper function.
"""

from enum import Enum
from typing import TypeVar, overload

import numpy as np
import scipp as sc
from scipp.core.concepts import irreducible_mask

T = TypeVar("T", bound=sc.Variable | sc.DataArray)


UncertaintyBroadcastMode = Enum(
    'UncertaintyBroadcastMode', ['drop', 'upper_bound', 'fail']
)
"""
Mode for broadcasting uncertainties.

- `drop`: Drop variances if the data is broadcasted.
- `upper_bound`: Compute an upper bound for the variances.
- `fail`: Do not broadcast, simply return the input data.

See https://doi.org/10.3233/JNR-220049 for context.
"""


@overload
def broadcast_with_upper_bound_variances(
    data: sc.Variable, /, *, prototype: sc.DataArray | sc.Variable
) -> sc.Variable:
    pass


@overload
def broadcast_with_upper_bound_variances(
    data: sc.DataArray, /, *, prototype: sc.DataArray | sc.Variable
) -> sc.DataArray:
    pass


def broadcast_with_upper_bound_variances(
    data: sc.Variable | sc.DataArray, /, *, prototype: sc.DataArray | sc.Variable
) -> sc.Variable | sc.DataArray:
    """
    Compute an upper bound for the variances of the broadcasted data.

    The variances of the broadcasted data are computed by scaling the variances of the
    input data by the volume of the new subspace. The volume of the new subspace is
    computed as the product of the sizes of the new dimensions. In the case of an
    event-data prototype the events are counted.

    Parameters
    ----------
    data:
        The data to broadcast.
    prototype:
        Defines the new sizes (dims and shape). If present, masks are used to exclude
        masked values from the variance computation.

    Returns
    -------
    :
        The data with the variances scaled by the volume of the new subspace.
    """
    if _no_variance_broadcast(data, prototype=prototype):
        return data
    for dim in prototype.dims:
        coord1 = None if isinstance(data, sc.Variable) else data.coords.get(dim)
        coord2 = (
            None if isinstance(prototype, sc.Variable) else prototype.coords.get(dim)
        )
        if coord1 is None or coord2 is None:
            if dim in data.dims:
                if data.sizes[dim] != prototype.sizes[dim]:
                    raise ValueError("Mismatching binning not supported in broadcast.")
            continue
        elif sc.identical(coord1, coord2):
            continue
        raise ValueError("Mismatching binning not supported in broadcast.")
    sizes = prototype.sizes
    mask = sc.scalar(False)
    if isinstance(prototype, sc.DataArray):
        if (irred := irreducible_mask(prototype, dim=sizes)) is not None:
            for dim in data.dims:
                if dim in irred.dims:
                    irred = irred.all(dim)
            mask = irred
    data = data.copy()
    sizes = {**sizes, **data.sizes}
    if prototype.bins is None:
        size = (~mask).sum().to(dtype='int64', copy=False)
        for dim, dim_size in sizes.items():
            if dim not in data.dims and dim not in mask.dims:
                size *= sc.index(dim_size)
    else:
        size = prototype.bins.size().sum(set(prototype.dims) - set(data.dims))
    scale = size.broadcast(sizes=sizes).to(dtype='float64')
    if not sc.identical(mask, sc.scalar(False)):
        # The masked values are not counted in the variance, so we set them to infinity.
        scale.values[mask.broadcast(sizes=sizes).values] = np.inf
    data = data.broadcast(sizes=sizes).copy()
    data.variances *= scale.values
    if prototype.bins is not None:
        # Note that we are not using event masks in the upper-bound computation. Less
        # than optimal, but simpler.
        if isinstance(data, sc.Variable):
            data = sc.bins_like(prototype, data)
        else:
            data.data = sc.bins_like(prototype, data.data)
    return data


@overload
def drop_variances_if_broadcast(
    data: sc.Variable, /, *, prototype: sc.DataArray | sc.Variable
) -> sc.Variable:
    pass


@overload
def drop_variances_if_broadcast(
    data: sc.DataArray, /, *, prototype: sc.DataArray | sc.Variable
) -> sc.DataArray:
    pass


def drop_variances_if_broadcast(
    data: sc.Variable | sc.DataArray, /, *, prototype: sc.DataArray | sc.Variable
) -> sc.Variable | sc.DataArray:
    """
    Drop variances if the data is broadcasted.

    Parameters
    ----------
    data:
        The data to broadcast.
    prototype:
        Defines the new sizes (dims and shape).

    Returns
    -------
    :
        The data without variances if the data is broadcasted.
    """
    if _no_variance_broadcast(data, prototype=prototype):
        return data
    return sc.values(data)


def _no_variance_broadcast(
    data: sc.Variable | sc.DataArray, /, *, prototype: sc.Variable | sc.DataArray
) -> bool:
    if data.bins is not None:
        raise ValueError("Cannot broadcast binned data.")
    if data.variances is None:
        return True
    if prototype.bins is not None:
        return False
    sizes = prototype.sizes
    return all(data.sizes.get(dim) == size for dim, size in sizes.items())


def _fail(
    data: sc.Variable | sc.DataArray, /, *, prototype: sc.Variable | sc.DataArray
) -> sc.Variable | sc.DataArray:
    # If there are variances, a subsequent broadcasting operation using Scipp will fail.
    # Do nothing here.
    return data


broadcasters = {
    UncertaintyBroadcastMode.drop: drop_variances_if_broadcast,
    UncertaintyBroadcastMode.upper_bound: broadcast_with_upper_bound_variances,
    UncertaintyBroadcastMode.fail: _fail,
}


def broadcast_uncertainties(
    data: sc.Variable | sc.DataArray,
    /,
    *,
    prototype: sc.DataArray | sc.Variable,
    mode: UncertaintyBroadcastMode,
) -> sc.Variable | sc.DataArray:
    """Broadcast uncertainties using the specified mode.

    Since Scipp raises an error when broadcasting data with variances, this function
    provides an explicit way to handle variances during broadcast operations.

    Parameters
    ----------
    data:
        Data with uncertainties to broadcast.
    prototype:
        Prototype defining the new sizes (dims and shape, or binned data sizes).
    mode:
        Selected broadcast mode.

    Returns
    -------
    :
        Data with broadcast uncertainties.
    """
    return broadcasters[mode](data, prototype=prototype)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import scipp as sc
from ipywidgets import FloatText, HBox, Label, Text, ValueWidget


class VectorWidget(HBox, ValueWidget):
    def __init__(self, name: str, variable: sc.Variable, components: str):
        super().__init__()

        style = {
            "layout": {"width": "130px"},
            "style": {"description_width": "initial"},
        }
        self.fields = {
            c: FloatText(
                description=f"{c} =", value=getattr(variable.fields, c).value, **style
            )
            for c in components
        }
        self.fields["unit"] = Text(
            description="unit:", value=str(variable.unit), **style
        )
        self.children = [Label(value=f"{name}: "), *list(self.fields.values())]

    @property
    def value(self):
        return sc.vector(
            value=[
                self.fields['x'].value,
                self.fields['y'].value,
                self.fields.get('z', sc.scalar(0.0)).value,
            ],
            unit=self.fields['unit'].value,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

"""NeXus loaders."""

from collections.abc import Mapping
from contextlib import AbstractContextManager, nullcontext
from dataclasses import dataclass
from math import prod
from typing import cast

import scipp as sc
import scippnexus as snx

from ..logging import get_logger
from .types import FilePath, NeXusEntryName, NeXusFile, NeXusGroup, NeXusLocationSpec


class NoNewDefinitionsType: ...


NoNewDefinitions = NoNewDefinitionsType()


def load_component(
    location: NeXusLocationSpec,
    *,
    nx_class: type[snx.NXobject],
    definitions: Mapping | None | NoNewDefinitionsType = NoNewDefinitions,
) -> sc.DataGroup:
    file_path = location.filename
    selection = location.selection
    entry_name = location.entry_name
    group_name = location.component_name
    with _open_nexus_file(file_path, definitions=definitions) as f:
        entry = _unique_child_group(f, snx.NXentry, entry_name)
        if nx_class is snx.NXsample:
            instrument = entry
        else:
            instrument = _unique_child_group(entry, snx.NXinstrument, None)
        component = _unique_child_group(instrument, nx_class, group_name)
        loaded = cast(sc.DataGroup, component[selection])
        loaded['nexus_component_name'] = component.name.split('/')[-1]
    return loaded


def compute_component_position(dg: sc.DataGroup) -> sc.DataGroup:
    # In some downstream packages we use some of the Nexus components which attempt
    # to compute positions without having actual Nexus data defining depends_on chains.
    # We assume positions have been set in the non-Nexus input somehow and return early.
    if 'depends_on' not in dg:
        return dg
    transform_out_name = 'transform'
    if transform_out_name in dg:
        raise RuntimeError(
            f"Loaded data contains an item '{transform_out_name}' but we want to "
            "store the combined NeXus transformations under that name."
        )
    position_out_name = 'position'
    if position_out_name in dg:
        raise RuntimeError(
            f"Loaded data contains an item '{position_out_name}' but we want to "
            "store the computed positions under that name."
        )
    return snx.compute_positions(
        dg, store_position=position_out_name, store_transform=transform_out_name
    )


def _open_nexus_file(
    file_path: FilePath | NeXusFile | NeXusGroup,
    definitions: Mapping | None | NoNewDefinitionsType = NoNewDefinitions,
) -> AbstractContextManager:
    if isinstance(file_path, getattr(NeXusGroup, '__supertype__', type(None))):
        if (
            definitions is not NoNewDefinitions
            and definitions != file_path._definitions
        ):
            raise ValueError(
                "Cannot apply new definitions to open nexus file or nexus group."
            )
        return nullcontext(file_path)
    if definitions is NoNewDefinitions:
        return snx.File(file_path)
    return snx.File(file_path, definitions=definitions)


def _unique_child_group(
    group: snx.Group, nx_class: type[snx.NXobject], name: str | None
) -> snx.Group:
    if name is not None:
        child = group[name]
        if isinstance(child, snx.Field):
            raise ValueError(
                f"Expected a NeXus group as item '{name}' but got a field."
            )
        if child.nx_class != nx_class:
            raise ValueError(
                f"The NeXus group '{name}' was expected to be a "
                f'{nx_class} but is a {child.nx_class}.'
            )
        return child

    children = group[nx_class]
    if len(children) != 1:
        raise ValueError(
            f"Expected exactly one {nx_class.__name__} group '{group.name}', "
            f"got {len(children)}"
        )
    return next(iter(children.values()))  # type: ignore[return-value]


def _contains_nx_class(group: snx.Group, nx_class: type[snx.NXobject]) -> bool:
    # See https://github.com/scipp/scippnexus/issues/241
    try:
        return bool(group[nx_class])
    except KeyError:
        # This does not happen with the current implementation in ScippNexus.
        # The fallback is here to future-proof this function.
        return False


def extract_signal_data_array(dg: sc.DataGroup) -> sc.DataArray:
    event_data_arrays = {
        key: value
        for key, value in dg.items()
        if isinstance(value, sc.DataArray) and value.bins is not None
    }
    # Transformations and thus the position can be time dependent, in which case they
    # are stored as DataArray instead of Variable. We do not want to select these.
    histogram_data_arrays = {
        key: value
        for key, value in dg.items()
        if isinstance(value, sc.DataArray)
        and value.bins is None
        and key not in ('position', 'transform')
    }
    if (array := _select_unique_array(event_data_arrays, 'event')) is not None:
        if histogram_data_arrays:
            get_logger().info(
                "Selecting event data '%s' in favor of histogram data {%s}",
                next(iter(event_data_arrays.keys())),
                ', '.join(f"'{k}'" for k in histogram_data_arrays),
            )
        return array

    if (array := _select_unique_array(histogram_data_arrays, 'histogram')) is not None:
        return array

    raise ValueError(
        "Raw data loaded from NeXus does not contain events or a histogram. "
        "Expected to find a data array, "
        f"but the data only contains {set(dg.keys())}"
    )


def _select_unique_array(
    arrays: dict[str, sc.DataArray], mapping_name: str
) -> sc.DataArray | None:
    if not arrays:
        return None
    if len(arrays) > 1:
        raise ValueError(
            f"Raw data loaded from NeXus contains more than one {mapping_name} "
            "data array. Cannot uniquely identify the data to extract. "
            f"Got {mapping_name} items {set(arrays.keys())}"
        )
    return next(iter(arrays.values()))


def _to_snx_selection(selection, *, for_events: bool) -> snx.typing.ScippIndex:
    if selection == slice(None, None):
        return ()
    if isinstance(selection, slice):
        if for_events:
            return {'event_time_zero': selection}
        return {'time': selection}
    return selection


def load_data(
    file_path: FilePath | NeXusFile | NeXusGroup,
    selection: snx.typing.ScippIndex | slice = (),
    *,
    entry_name: NeXusEntryName | None = None,
    component_name: str,
    definitions: Mapping | NoNewDefinitionsType = NoNewDefinitions,
) -> sc.DataArray:
    """Load data of a detector or monitor from a NeXus file.

    Loads either event data from an ``NXevent_data`` group or histogram
    data from an ``NXdata`` group depending on which ``group`` contains.
    Event data is grouped by ``'event_time_zero'`` as in the NeXus file.
    Histogram data is returned as encoded in the file.

    Parameters
    ----------
    file_path:
        Indicates where to load data from.
        One of:

        - Path to a NeXus file on disk.
        - File handle or buffer for reading binary data.
        - A ScippNexus group of the root of a NeXus file.
    selection:
        Select which aprt of the data to load.
        By default, load all data.
        Supports anything that ScippNexus supports.
    component_name:
        Name of the NXdetector or NXmonitor containing the NXevent_data to load.
        Must be a group in an instrument group in the entry (see below).
    entry_name:
        Name of the entry that contains the detector.
        If ``None``, the entry will be located based
        on its NeXus class, but there cannot be more than 1.
    definitions:
        Definitions used by scippnexus loader, see :py:`scippnexus.File`
        for documentation.

    Returns
    -------
    :
        Data array with events or a histogram.
    """
    with _open_nexus_file(file_path, definitions=definitions) as f:
        entry = _unique_child_group(f, snx.NXentry, entry_name)
        instrument = _unique_child_group(entry, snx.NXinstrument, None)
        component = instrument[component_name]
        if _contains_nx_class(component, snx.NXevent_data):
            data = _unique_child_group(component, snx.NXevent_data, None)
            sel = _to_snx_selection(selection, for_events=True)
        elif _contains_nx_class(component, snx.NXdata):
            data = _unique_child_group(component, snx.NXdata, None)
            sel = _to_snx_selection(selection, for_events=False)
        else:
            raise ValueError(
                f"NeXus group '{component.name}' contains neither "
                "NXevent_data nor NXdata."
            )

        return data[sel]


def group_event_data(
    *, event_data: sc.DataArray, detector_number: sc.Variable
) -> sc.DataArray:
    """Group event data by detector number.

    The detector_number variable also defines the output shape and dimension names.

    Parameters
    ----------
    event_data:
        Data array with events to group, as returned from :py:func:`load_event_data`.
    detector_number:
        Variable with detector numbers matching the `event_id` field of the event data.

    Returns
    -------
    :
        Data array with events grouped by detector number.
    """
    event_id = detector_number.flatten(to='event_id').copy()
    constituents = event_data.bins.constituents
    begin = constituents['begin']
    end = constituents['end']
    data = constituents['data'].copy(deep=False)
    if 'event_time_zero' in event_data.coords:
        data.coords['event_time_zero'] = sc.bins_like(
            event_data, fill_value=event_data.coords['event_time_zero']
        ).bins.constituents['data']
    # After loading raw NXevent_data it is guaranteed that the event table
    # is contiguous and that there is no masking. We can therefore use the
    # more efficient approach of binning from scratch instead of erasing the
    # 'event_time_zero' binning defined by NXevent_data. This sanity check should
    # therefore always pass unless some unusual modifications were performed.
    if (
        event_data.masks
        or begin[0] != sc.index(0)
        or end[-1] != sc.index(data.sizes[data.dim])
        or (begin[1:] != end[:-1]).any()
    ):
        raise ValueError("Grouping only implemented for contiguous data with no masks.")
    out = data.group(event_id).fold(dim='event_id', sizes=detector_number.sizes)
    out.coords['detector_number'] = out.coords.pop('event_id')
    return out


def _format_time(time: sc.Variable | None) -> str:
    if time is None:
        return 'None'
    return f"{time:c}" if time.dtype != 'datetime64' else str(time.value)


@dataclass
class NeXusDetectorInfo:
    name: str
    start_time: sc.Variable | None
    end_time: sc.Variable | None
    n_pulse: int | None
    n_pixel: int | None

    def __repr__(self) -> str:
        return (
            f"{self.name}: n_pulse={self.n_pulse}, n_pixel={self.n_pixel}, "
            f"start_time={_format_time(self.start_time)}, "
            f"end_time={_format_time(self.end_time)}"
        )


@dataclass
class NeXusMonitorInfo:
    name: str
    start_time: sc.Variable | None
    end_time: sc.Variable | None
    n_pulse: int | None

    def __repr__(self) -> str:
        return (
            f"{self.name}: n_pulse={self.n_pulse}, "
            f"start_time={_format_time(self.start_time)}, "
            f"end_time={_format_time(self.end_time)}"
        )


@dataclass
class NeXusFileInfo:
    detectors: dict[str, NeXusDetectorInfo]
    monitors: dict[str, NeXusMonitorInfo]

    @property
    def start_time(self) -> sc.Variable | None:
        times = [
            comp.start_time
            for comp in (*self.detectors.values(), *self.monitors.values())
            if comp.start_time is not None
        ]
        return sc.reduce(times).min() if times else None

    @property
    def end_time(self) -> sc.Variable | None:
        times = [
            comp.end_time
            for comp in (*self.detectors.values(), *self.monitors.values())
            if comp.end_time is not None
        ]
        return sc.reduce(times).max() if times else None

    def __repr__(self) -> str:
        s = "NeXusFileInfo(\n"
        s += "  Detectors:\n"
        s += "\n".join(f"    {det}" for det in self.detectors.values())
        s += "\n  Monitors:\n"
        s += "\n".join(f"    {mon}" for mon in self.monitors.values())
        s += ')'
        return s


def _parse_name(group: snx.Group) -> str:
    return group.name.split('/')[-1]


def _parse_pixel_count(group: snx.Group) -> int | None:
    if (detector_number := group.get('detector_number')) is not None:
        return prod(detector_number.shape)


def _parse_pulse_count(group: snx.Group) -> int | None:
    try:
        events = _unique_child_group(group, snx.NXevent_data, None)
    except ValueError:
        return None
    try:
        return events['event_index'].shape[0]
    except KeyError:
        return None


def _get_start_time(group: snx.Group) -> sc.Variable | None:
    try:
        events = _unique_child_group(group, snx.NXevent_data, None)
    except ValueError:
        return None
    try:
        return events['event_time_zero'][0]
    except KeyError:
        return None


def _get_end_time(group: snx.Group) -> sc.Variable | None:
    try:
        events = _unique_child_group(group, snx.NXevent_data, None)
    except ValueError:
        return None
    try:
        return events['event_time_zero'][-1]
    except KeyError:
        return None


def _parse_detector(group: snx.Group) -> NeXusDetectorInfo:
    return NeXusDetectorInfo(
        name=_parse_name(group),
        start_time=_get_start_time(group),
        end_time=_get_end_time(group),
        n_pulse=_parse_pulse_count(group),
        n_pixel=_parse_pixel_count(group),
    )


def _parse_monitor(group: snx.Group) -> NeXusMonitorInfo:
    return NeXusMonitorInfo(
        name=_parse_name(group),
        start_time=_get_start_time(group),
        end_time=_get_end_time(group),
        n_pulse=_parse_pulse_count(group),
    )


def read_nexus_file_info(file_path: FilePath | NeXusFile | NeXusGroup) -> NeXusFileInfo:
    """Opens and inspects a NeXus file, returning a summary of its contents."""
    with _open_nexus_file(file_path) as f:
        entry = _unique_child_group(f, snx.NXentry, None)
        instrument = _unique_child_group(entry, snx.NXinstrument, None)
        detectors = {}
        monitors = {}
        for name, obj in instrument.items():
            if not isinstance(obj, snx.Group):
                continue
            if obj.nx_class == snx.NXdetector:
                detectors[name] = _parse_detector(obj)
            elif obj.nx_class == snx.NXmonitor:
                monitors[name] = _parse_monitor(obj)

        return NeXusFileInfo(detectors=detectors, monitors=monitors)



import os
import tempfile

import h5py
import numpy as np
import pytest

from ess.reduce.scripts.grow_nexus import grow_nexus_file


@pytest.fixture()
def nexus_file():
    with tempfile.TemporaryDirectory() as tmp:
        path = os.path.join(tmp, 'test.nxs')
        with h5py.File(path, 'a') as hf:
            entry = hf.create_group('entry')
            entry.attrs['NX_class'] = 'NXentry'

            instrument = entry.create_group('instrument')
            instrument.attrs['NX_class'] = 'NXinstrument'

            for group, nxclass in (
                ('detector', 'NXdetector'),
                ('monitor', 'NXmonitor'),
            ):
                detector = instrument.create_group(group)
                detector.attrs['NX_class'] = nxclass

                event_data = detector.create_group('event_data')
                event_data.attrs['NX_class'] = 'NXevent_data'

                event_data.create_dataset(
                    'event_index',
                    data=np.array([2, 4, 6]),
                    maxshape=(None,),
                    chunks=True,
                )
                event_data.create_dataset(
                    'event_time_zero',
                    data=np.array([0, 1, 2]),
                    maxshape=(None,),
                    chunks=True,
                )
                event_data.create_dataset(
                    'event_id',
                    data=np.array([0, 1, 2, 0, 1, 2]),
                    maxshape=(None,),
                    chunks=True,
                )
                event_data.create_dataset(
                    'event_time_offset',
                    data=np.array([1, 2, 1, 2, 1, 2]),
                    maxshape=(None,),
                    chunks=True,
                )

            yield path


@pytest.mark.parametrize('monitor_scale', [1, 2, None])
@pytest.mark.parametrize('detector_scale', [1, 2])
def test_grow_nexus(nexus_file, detector_scale, monitor_scale):
    grow_nexus_file(
        filename=nexus_file, detector_scale=detector_scale, monitor_scale=monitor_scale
    )

    monitor_scale = monitor_scale if monitor_scale is not None else detector_scale

    with h5py.File(nexus_file, 'r') as f:
        for detector, scale in zip(
            ('detector', 'monitor'), (detector_scale, monitor_scale), strict=True
        ):
            np.testing.assert_equal(
                [scale * i for i in [2, 4, 6]],
                f[f'entry/instrument/{detector}/event_data/event_index'][()],
            )
            np.testing.assert_equal(
                scale * [0, 1, 2, 0, 1, 2],
                f[f'entry/instrument/{detector}/event_data/event_id'][()],
            )
            np.testing.assert_equal(
                scale * [1, 2, 1, 2, 1, 2],
                f[f'entry/instrument/{detector}/event_data/event_time_offset'][()],
            )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from collections.abc import MutableMapping
from dataclasses import dataclass
from enum import Enum
from typing import Generic, TypeVar

import scipp as sc
from sciline._utils import key_name
from sciline.typing import Key

T = TypeVar('T')
C = TypeVar('C', bound='Parameter')


class KeepDefaultType:
    pass


keep_default = KeepDefaultType()


@dataclass
class Parameter(Generic[T]):
    """Interface between workflow parameters and widgets.

    .. versionadded:: RELEASE_PLACEHOLDER
    """

    name: str
    description: str
    default: T
    optional: bool = False
    """If True, widget has radio buttons switch between "None" and param widget."""
    switchable: bool = False
    """If True, widget has checkbox to enable/disable parameter."""

    def with_default(self: C, default: T | KeepDefaultType = keep_default) -> C:
        if default == keep_default:
            return self
        # TODO I think some subclasses currently cannot be instantiated with this method
        return type(self)(
            self.name, self.description, default, self.optional, self.switchable
        )

    @classmethod
    def from_type(
        cls: type[C],
        t: type[T],
        default: T | None = None,
        optional: bool = False,
        switchable: bool = False,
    ) -> C:
        # TODO __doc__ not correct when using NewType
        # TODO __doc__ not correct when using Generic
        # use sciline type->string helper
        return cls(
            name=key_name(t),
            description=t.__doc__,
            default=default,
            optional=optional,
            switchable=switchable,
        )


@dataclass(kw_only=True)
class ParamWithOptions(Parameter[T]):
    options: Enum

    @classmethod
    def from_enum(cls: type[C], t: type[T], default: T) -> C:
        return cls(name=str(t), description=t.__doc__, options=t, default=default)


@dataclass
class FilenameParameter(Parameter[str]):
    """Widget for entering a filename or selecting one in a file dialog."""

    # TODO need specifics for different file types, nexus, ...


@dataclass
class MultiFilenameParameter(Parameter[tuple[str, ...]]):
    """Widget for entering multiple filenames or selecting multiple in a file dialog."""


@dataclass(kw_only=True)
class BinEdgesParameter(Parameter[sc.Variable]):
    """Widget for entering bin edges."""

    dim: str
    start: float | None = None
    stop: float | None = None
    nbins: int = 1
    unit: str | None = "undefined"  # If "undefined", the unit is deduced from the dim
    log: bool = False

    def __init__(
        self,
        t: type[T],
        dim: str,
        start: float | None = None,
        stop: float | None = None,
        nbins: int = 1,
        unit: str | None = "undefined",
        log: bool = False,
    ):
        self.dim = dim
        self.start = start
        self.stop = stop
        self.nbins = nbins
        self.unit = unit
        self.log = log
        super().__init__(name=key_name(t), description=t.__doc__, default=None)


@dataclass
class BooleanParameter(Parameter[bool]):
    pass


@dataclass
class StringParameter(Parameter[str]):
    pass


@dataclass(kw_only=True)
class ParamWithBounds(Parameter[T]):
    bounds: tuple[T, T]


@dataclass(kw_only=True)
class ScalarParameter(Parameter[T]):
    """Fixed unit displayed in widget"""

    unit: str


@dataclass(kw_only=True)
class ScalarParamWithUnitOptions(Parameter[T]):
    """User can select between compatible units"""

    unit_options: list[str]


@dataclass(kw_only=True)
class Vector2dParameter(Parameter[sc.Variable]):
    """Widget for entering a 2d vector."""


@dataclass(kw_only=True)
class Vector3dParameter(Parameter[sc.Variable]):
    """Widget for entering a 3d vector."""


class ParameterRegistry(MutableMapping):
    def __init__(self):
        self._parameters = {}

    def __getitem__(self, key: Key) -> Parameter:
        return self._parameters[key]

    def __setitem__(self, key: Key, value: Parameter):
        self._parameters[key] = value

    def __delitem__(self, key: Key):
        del self._parameters[key]

    def __iter__(self):
        return iter(self._parameters)

    def __len__(self) -> int:
        return len(self._parameters)


parameter_registry = ParameterRegistry()


parameter_mappers = {}



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

"""Workflow and workflow components for interacting with NeXus files."""

from collections.abc import Sequence
from copy import deepcopy
from typing import Any

import networkx as nx
import sciline
import scipp as sc
import scippnexus as snx
from scipp.constants import g
from scipp.core import label_based_index_to_positional_index

from . import _nexus_loader as nexus
from .types import (
    CalibratedBeamline,
    CalibratedDetector,
    CalibratedMonitor,
    Component,
    DetectorBankSizes,
    DetectorData,
    DetectorPositionOffset,
    Filename,
    GravityVector,
    MonitorData,
    MonitorPositionOffset,
    MonitorType,
    NeXusClass,
    NeXusComponent,
    NeXusComponentLocationSpec,
    NeXusData,
    NeXusDataLocationSpec,
    NeXusFileSpec,
    NeXusName,
    NeXusTransformation,
    NeXusTransformationChain,
    Position,
    PreopenNeXusFile,
    RunType,
    TimeInterval,
    UniqueComponent,
)

origin = sc.vector([0, 0, 0], unit="m")
"""The origin, used as default sample position."""
no_offset = sc.vector([0, 0, 0], unit="m")
"""Offset that does not change the position."""


def file_path_to_file_spec(
    filename: Filename[RunType], preopen: PreopenNeXusFile
) -> NeXusFileSpec[RunType]:
    return NeXusFileSpec[RunType](
        snx.File(filename, definitions=definitions) if preopen else filename
    )


def no_monitor_position_offset() -> MonitorPositionOffset[RunType, MonitorType]:
    return MonitorPositionOffset[RunType, MonitorType](no_offset)


def no_detector_position_offset() -> DetectorPositionOffset[RunType]:
    return DetectorPositionOffset[RunType](no_offset)


def full_time_interval() -> TimeInterval[RunType]:
    """Select all neutron pulses in the data."""
    return TimeInterval[RunType](slice(None, None))


def gravity_vector_neg_y() -> GravityVector:
    """
    Gravity vector for default instrument coordinate system where y is up.
    """
    return GravityVector(sc.vector(value=[0, -1, 0]) * g)


def component_spec_by_name(
    filename: NeXusFileSpec[RunType], name: NeXusName[Component]
) -> NeXusComponentLocationSpec[Component, RunType]:
    """
    Create a location spec for a component group in a NeXus file.

    Parameters
    ----------
    filename:
        NeXus file to use for the location spec.
    name:
        Name of the component group.
    """
    return NeXusComponentLocationSpec[Component, RunType](
        filename=filename.value, component_name=name
    )


def unique_component_spec(
    filename: NeXusFileSpec[RunType],
) -> NeXusComponentLocationSpec[UniqueComponent, RunType]:
    """
    Create a location spec for a unique component group in a NeXus file.

    Parameters
    ----------
    filename:
        NeXus file to use for the location spec.
    """
    return NeXusComponentLocationSpec[UniqueComponent, RunType](filename=filename.value)


def data_by_name(
    filename: NeXusFileSpec[RunType],
    name: NeXusName[Component],
    selection: TimeInterval[RunType],
) -> NeXusDataLocationSpec[Component, RunType]:
    """
    Create a location spec for monitor or detector data in a NeXus file.

    Parameters
    ----------
    filename:
        NeXus file to use for the location spec.
    name:
        Name of the monitor or detector group.
    selection:
        Time range (start and stop as a Python slice object).
    """
    return NeXusDataLocationSpec[Component, RunType](
        filename=filename.value, component_name=name, selection=selection.value
    )


def load_nexus_sample(
    location: NeXusComponentLocationSpec[snx.NXsample, RunType],
) -> NeXusComponent[snx.NXsample, RunType]:
    """
    Load a NeXus sample group from a file.

    If there is no sample group in the file, an empty group is returned. This should
    not happen, but handling it gracefully makes testing and working with
    pre-production files easier. There should be little harm in returning an empty
    group. Subsequent extract of the sample position will then default to the origin.

    Parameters
    ----------
    location:
        Location spec for the sample group.
    """
    try:
        dg = nexus.load_component(location, nx_class=snx.NXsample)
    except ValueError:
        dg = sc.DataGroup()
    if 'depends_on' not in dg:
        dg['depends_on'] = snx.TransformationChain(parent='', value='.')
    return NeXusComponent[snx.NXsample, RunType](dg)


def nx_class_for_monitor() -> NeXusClass[MonitorType]:
    return NeXusClass[MonitorType](snx.NXmonitor)


def nx_class_for_detector() -> NeXusClass[snx.NXdetector]:
    return NeXusClass[snx.NXdetector](snx.NXdetector)


def nx_class_for_source() -> NeXusClass[snx.NXsource]:
    return NeXusClass[snx.NXsource](snx.NXsource)


def nx_class_for_sample() -> NeXusClass[snx.NXsample]:
    return NeXusClass[snx.NXsample](snx.NXsample)


def load_nexus_component(
    location: NeXusComponentLocationSpec[Component, RunType],
    nx_class: NeXusClass[Component],
) -> NeXusComponent[Component, RunType]:
    """
    Load a NeXus component group from a file.

    When loading a detector or monitor, event data is replaced by placeholders.

    As the event data can be large and is not needed at this stage, it is replaced by
    a placeholder. A placeholder is used to allow for returning a scipp.DataArray, which
    is what most downstream code will expect.
    Currently the placeholder is the detector number (for detectors) or a size-0 array
    (for monitors), but this may change in the future.

    The returned object is a scipp.DataGroup, as it may contain additional information
    about the detector that cannot be represented as a single scipp.DataArray. Most
    downstream code will only be interested in the contained scipp.DataArray so this
    needs to be extracted. However, other processing steps may require the additional
    information, so it is kept in the DataGroup.

    Parameters
    ----------
    location:
        Location spec for the source group.
    nx_class:
        NX_class to identify the component.
    """
    return NeXusComponent[Component, RunType](
        nexus.load_component(location, nx_class=nx_class, definitions=definitions)
    )


def load_nexus_data(
    location: NeXusDataLocationSpec[Component, RunType],
) -> NeXusData[Component, RunType]:
    """
    Load event or histogram data from a NeXus detector group.

    Parameters
    ----------
    location:
        Location spec for the detector group.
    """
    return NeXusData[Component, RunType](
        nexus.load_data(
            file_path=location.filename,
            entry_name=location.entry_name,
            selection=location.selection,
            component_name=location.component_name,
        )
    )


def get_transformation_chain(
    detector: NeXusComponent[Component, RunType],
) -> NeXusTransformationChain[Component, RunType]:
    """
    Extract the transformation chain from a NeXus detector group.

    Parameters
    ----------
    detector:
        NeXus detector group.
    """
    chain = detector['depends_on']
    return NeXusTransformationChain[Component, RunType](chain)


def _time_filter(transform: sc.DataArray) -> sc.Variable:
    if transform.ndim == 0 or transform.sizes == {'time': 1}:
        return transform.data.squeeze()
    raise ValueError(
        f"Transform is time-dependent: {transform}, but no filter is provided."
    )


def to_transformation(
    chain: NeXusTransformationChain[Component, RunType], interval: TimeInterval[RunType]
) -> NeXusTransformation[Component, RunType]:
    """
    Convert transformation chain into a single transformation matrix.

    If one or more transformations in the chain are time-dependent, the time interval
    is used to select a specific time point. If the interval is not a single time point,
    an error is raised. This may be extended in the future to a more sophisticated
    mechanism, e.g., averaging over the interval to remove noise.

    Parameters
    ----------
    chain:
        Transformation chain.
    interval:
        Time interval to select from the transformation chain.
    """

    chain = deepcopy(chain)
    for t in chain.transformations.values():
        if t.sizes == {} or not isinstance(t.value, sc.DataArray):
            continue
        start = interval.value.start
        stop = interval.value.stop
        if isinstance(start, sc.Variable) or isinstance(stop, sc.Variable):
            # NXlog entries are generally interpreted as the previous value being valid
            # until the next entry. We therefore need to select the previous value, and
            # any index after the last entry refers to the last entry, i.e., there is no
            # "end" time in the files. We add a dummy end so we can use Scipp's label-
            # based indexing for histogram data.
            time = t.value.coords['time']
            # Add 1000 days as a dummy end time. This is hopefully long enough to cover
            # all reasonable use cases.
            delta = sc.scalar(24_000, unit='hours', dtype='int64').to(unit=time.unit)
            time = sc.concat([time, time[-1] + delta], 'time')
            idx = label_based_index_to_positional_index(
                sizes=t.sizes, coord=time, index=interval.value
            )
            t.value = _time_filter(t.value[idx])
        else:
            t.value = _time_filter(t.value['time', interval.value])

    return NeXusTransformation[Component, RunType].from_chain(chain)


def compute_position(
    transformation: NeXusTransformation[Component, RunType],
) -> Position[Component, RunType]:
    """Compute the position of a component from a transformation matrix."""
    return Position[Component, RunType](transformation.value * origin)


def get_calibrated_detector(
    detector: NeXusComponent[snx.NXdetector, RunType],
    *,
    transform: NeXusTransformation[snx.NXdetector, RunType],
    # Strictly speaking we could apply an offset by modifying the transformation chain,
    # using a more generic implementation. However, this may in general require
    # extending the chain and it is currently not clear if that is desirable. As far as
    # I am aware the offset is currently mainly used for handling files from other
    # facilities and it is not clear if it is needed for ESS data and should be kept at
    # all.
    offset: DetectorPositionOffset[RunType],
    bank_sizes: DetectorBankSizes,
) -> CalibratedDetector[RunType]:
    """
    Extract the data array corresponding to a detector's signal field.

    The returned data array includes coords and masks pertaining directly to the
    signal values array, but not additional information about the detector. The
    data array is reshaped to the logical detector shape, which by folding the data
    array along the detector_number dimension.

    Parameters
    ----------
    detector:
        NeXus detector group.
    offset:
        Offset to add to the detector position.
    bank_sizes:
        Dictionary of detector bank sizes.
    """
    da = nexus.extract_signal_data_array(detector)
    if (
        sizes := (bank_sizes or {}).get(detector.get('nexus_component_name'))
    ) is not None:
        da = da.fold(dim="detector_number", sizes=sizes)
    # Note: We apply offset as early as possible, i.e., right in this function
    # the detector array from the raw loader NeXus group, to prevent a source of bugs.
    # If the NXdetector in the file is not 1-D, we want to match the order of dims.
    # zip_pixel_offsets otherwise yields a vector with dimensions in the order given
    # by the x/y/z offsets.
    offsets = snx.zip_pixel_offsets(da.coords).transpose(da.dims).copy()
    # We use the unit of the offsets as this is likely what the user expects.
    position = transform.value.to(unit=offsets.unit) * offsets
    return CalibratedDetector[RunType](
        da.assign_coords(position=position + offset.to(unit=position.unit))
    )


def assemble_beamline(
    detector: CalibratedDetector[RunType],
    source_position: Position[snx.NXsource, RunType],
    sample_position: Position[snx.NXsample, RunType],
    gravity: GravityVector,
) -> CalibratedBeamline[RunType]:
    """
    Add beamline information (gravity vector, source- and sample-position) to detector.

    This is performed separately and after :py:func:`get_calibrated_detector` to avoid
    as false dependency of, e.g., the reshaped detector numbers on the sample position.
    The latter can change during a run, e.g., for a rotating sample. The detector
    numbers might be used, e.g., to mask certain detector pixels, and should not depend
    on the sample position.

    Parameters
    ----------
    detector:
        NeXus detector group.
    source_position:
        Position of the neutron source.
    sample_position:
        Position of the sample.
    gravity:
        Gravity vector.
    """
    return CalibratedBeamline[RunType](
        detector.assign_coords(
            source_position=source_position,
            sample_position=sample_position,
            gravity=gravity,
        )
    )


def assemble_detector_data(
    detector: CalibratedBeamline[RunType],
    event_data: NeXusData[snx.NXdetector, RunType],
) -> DetectorData[RunType]:
    """
    Assemble a detector data array with event data.

    Also adds variances to the event data if they are missing.

    Parameters
    ----------
    detector:
        Calibrated detector data array.
    event_data:
        Event data array.
    """
    grouped = nexus.group_event_data(
        event_data=event_data, detector_number=detector.coords['detector_number']
    )
    return DetectorData[RunType](
        _add_variances(grouped)
        .assign_coords(detector.coords)
        .assign_masks(detector.masks)
    )


def get_calibrated_monitor(
    monitor: NeXusComponent[MonitorType, RunType],
    offset: MonitorPositionOffset[RunType, MonitorType],
    source_position: Position[snx.NXsource, RunType],
) -> CalibratedMonitor[RunType, MonitorType]:
    """
    Extract the data array corresponding to a monitor's signal field.

    The returned data array includes coords pertaining directly to the
    signal values array, but not additional information about the monitor.

    Parameters
    ----------
    monitor:
        NeXus monitor group.
    offset:
        Offset to add to the monitor position.
    source_position:
        Position of the neutron source.
    """
    monitor = nexus.compute_component_position(monitor)
    return CalibratedMonitor[RunType, MonitorType](
        nexus.extract_signal_data_array(monitor).assign_coords(
            position=monitor['position'] + offset.to(unit=monitor['position'].unit),
            source_position=source_position,
        )
    )


def assemble_monitor_data(
    monitor: CalibratedMonitor[RunType, MonitorType],
    data: NeXusData[MonitorType, RunType],
) -> MonitorData[RunType, MonitorType]:
    """
    Assemble a monitor data array with event data.

    Also adds variances to the event data if they are missing.

    Parameters
    ----------
    monitor:
        Calibrated monitor data array.
    data:
        Data array with neutron counts.
    """
    da = data.assign_coords(monitor.coords).assign_masks(monitor.masks)
    return MonitorData[RunType, MonitorType](_add_variances(da))


def _drop(
    children: dict[str, snx.Field | snx.Group], classes: tuple[snx.NXobject, ...]
) -> dict[str, snx.Field | snx.Group]:
    return {
        name: child
        for name, child in children.items()
        if not (isinstance(child, snx.Group) and (child.nx_class in classes))
    }


class _StrippedDetector(snx.NXdetector):
    """Detector definition without large geometry or event data for ScippNexus.

    Drops NXoff_geometry and NXevent_data groups, data is replaced by detector_number.
    """

    def __init__(
        self, attrs: dict[str, Any], children: dict[str, snx.Field | snx.Group]
    ):
        children = _drop(children, (snx.NXoff_geometry, snx.NXevent_data))
        children['data'] = children['detector_number']
        super().__init__(attrs=attrs, children=children)


class _DummyField:
    """Dummy field that can replace snx.Field in NXmonitor."""

    def __init__(self):
        self.attrs = {}
        self.sizes = {'event_time_zero': 0}
        self.dims = ('event_time_zero',)
        self.shape = (0,)

    def __getitem__(self, key: Any) -> sc.Variable:
        return sc.empty(dims=self.dims, shape=self.shape, unit=None)


class _StrippedMonitor(snx.NXmonitor):
    """Monitor definition without event data for ScippNexus.

    Drops NXevent_data group, data is replaced by a dummy field.
    """

    def __init__(
        self, attrs: dict[str, Any], children: dict[str, snx.Field | snx.Group]
    ):
        children = _drop(children, (snx.NXevent_data,))
        children['data'] = _DummyField()
        super().__init__(attrs=attrs, children=children)


def _add_variances(da: sc.DataArray) -> sc.DataArray:
    out = da.copy(deep=False)
    if out.bins is not None:
        content = out.bins.constituents['data']
        if content.variances is None:
            content.variances = content.values
    return out


definitions = snx.base_definitions()
definitions["NXdetector"] = _StrippedDetector
definitions["NXmonitor"] = _StrippedMonitor


_common_providers = (
    gravity_vector_neg_y,
    file_path_to_file_spec,
    full_time_interval,
    component_spec_by_name,
    unique_component_spec,  # after component_spec_by_name, partially overrides
    get_transformation_chain,
    to_transformation,
    compute_position,
    load_nexus_data,
    load_nexus_component,
    data_by_name,
    nx_class_for_detector,
    nx_class_for_monitor,
    nx_class_for_source,
    nx_class_for_sample,
)

_monitor_providers = (
    no_monitor_position_offset,
    get_calibrated_monitor,
    assemble_monitor_data,
)

_detector_providers = (
    no_detector_position_offset,
    load_nexus_sample,
    get_calibrated_detector,
    assemble_beamline,
    assemble_detector_data,
)


def LoadMonitorWorkflow() -> sciline.Pipeline:
    """Generic workflow for loading monitor data from a NeXus file."""
    wf = sciline.Pipeline((*_common_providers, *_monitor_providers))
    wf[PreopenNeXusFile] = PreopenNeXusFile(False)
    return wf


def LoadDetectorWorkflow() -> sciline.Pipeline:
    """Generic workflow for loading detector data from a NeXus file."""
    wf = sciline.Pipeline((*_common_providers, *_detector_providers))
    wf[DetectorBankSizes] = DetectorBankSizes({})
    wf[PreopenNeXusFile] = PreopenNeXusFile(False)
    return wf


def GenericNeXusWorkflow(
    *,
    run_types: Sequence[sciline.typing.Key] | None = None,
    monitor_types: Sequence[sciline.typing.Key] | None = None,
) -> sciline.Pipeline:
    """
    Generic workflow for loading detector and monitor data from a NeXus file.

    Parameters
    ----------
    run_types:
        List of run types to include in the workflow. If not provided, all run types
        are included. It is recommended to specify run types to avoid creating very
        large workflows.
    monitor_types:
        List of monitor types to include in the workflow. If not provided, all monitor
        types are included. It is recommended to specify monitor types to avoid creating
        very large workflows.

    Returns
    -------
    :
        The workflow.
    """
    if monitor_types is not None and run_types is None:
        raise ValueError("run_types must be specified if monitor_types is specified")
    wf = sciline.Pipeline(
        (*_common_providers, *_monitor_providers, *_detector_providers)
    )
    wf[DetectorBankSizes] = DetectorBankSizes({})
    wf[PreopenNeXusFile] = PreopenNeXusFile(False)

    g = wf.underlying_graph
    ancestors = set()
    # DetectorData and MonitorData are the "final" outputs, so finding and removing all
    # their ancestors is what we need to strip unused run and monitor types.
    for rt in run_types or ():
        ancestors |= nx.ancestors(g, DetectorData[rt])
        ancestors.add(DetectorData[rt])
        for mt in monitor_types or ():
            ancestors |= nx.ancestors(g, MonitorData[rt, mt])
            ancestors.add(MonitorData[rt, mt])
    if run_types is not None:
        g.remove_nodes_from(set(g.nodes) - ancestors)
    return wf



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
"""Generators for "JSON" data from a NeXus file, for the purpose of testing."""

from collections.abc import Generator
from typing import Any

import scipp as sc


def _as_str_attr(value: Any, name: str) -> dict:
    val = str(value)
    return {"string_size": len(val), "type": "string", "name": name, "values": val}


def _variable_to_json(var: sc.Variable, name: str):
    attrs = []
    if var.dtype == sc.DType.datetime64:
        offset = var.min()
        attrs.append(_as_str_attr(offset.value, "offset"))
        var = var - offset
    if var.unit is not None:
        attrs.append(_as_str_attr(var.unit, "units"))
    return {
        "module": "dataset",
        "config": {
            "name": name,
            "values": var.values,
            "size": list(var.shape),
            "type": str(var.dtype),
        },
        "attributes": attrs,
    }


_event_index_0 = sc.array(dims=('dummy',), values=[0], unit=None)


def _event_data_pulse_to_json(pulse: sc.DataArray) -> dict:
    content = pulse.value.coords
    event_time_zero = sc.concat([pulse.coords['event_time_zero']], 'dummy')
    event_time_offset = content['event_time_offset']
    event_id = content.get('event_id')
    # I think we always have a pixel_id in the flatbuffer, so monitors just get ones?
    if event_id is None:
        event_id = sc.ones(sizes=event_time_offset.sizes, dtype='int32', unit=None)
    children = [
        _variable_to_json(event_time_zero, name='event_time_zero'),
        _variable_to_json(event_time_offset, name='event_time_offset'),
        _variable_to_json(event_id, name='event_id'),
        _variable_to_json(_event_index_0, name='event_index'),
    ]
    group = {
        "type": "group",
        "name": "events_0",
        "children": children,
        "attributes": [_as_str_attr("NXevent_data", name="NX_class")],
    }
    return group


def event_data_generator(data: sc.DataArray) -> Generator[dict, None, None]:
    """
    Generate JSON data for event data from a NeXus file.

    Parameters
    ----------
    data:
        A data array with event data, equivalent to what ScippNexus would load from an
        NXevent_data group in a NeXus file.

    Yields
    ------
    :
        A dict of data for a single event data pulse that can be wrapped in a
        :py:class:`ess.reduce.nexus.json_nexus.JSONGroup`.
    """
    for pulse in data:
        yield _event_data_pulse_to_json(pulse)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import scipp as sc
from ipywidgets import FloatText, GridBox, IntText, Label, ValueWidget


class LinspaceWidget(GridBox, ValueWidget):
    def __init__(self, dim: str, unit: str):
        super().__init__()

        self.fields = {
            'dim': Label(value=dim, description='dim'),
            'start': FloatText(description='start', value=0.0),
            'end': FloatText(description='end', value=1.0),
            'num': IntText(description='num', value=100),
            'unit': Label(description='unit', value=unit),
        }
        self.children = [
            Label(value="Select range:"),
            self.fields['dim'],
            self.fields['unit'],
            self.fields['start'],
            self.fields['end'],
            self.fields['num'],
        ]

    @property
    def value(self) -> sc.Variable:
        return sc.linspace(
            self.fields['dim'].value,
            self.fields['start'].value,
            self.fields['end'].value,
            self.fields['num'].value,
            unit=self.fields['unit'].value,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""This module provides tools for running workflows in a streaming fashion."""

from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any, Generic, TypeVar

import networkx as nx
import sciline
import scipp as sc

T = TypeVar('T')


def maybe_hist(value: T) -> T:
    """
    Convert value to a histogram if it is not already a histogram.

    This is the default pre-processing used by accumulators.

    Parameters
    ----------
    value:
        Value to be converted to a histogram.

    Returns
    -------
    :
        Histogram.
    """
    return value if value.bins is None else value.hist()


class Accumulator(ABC, Generic[T]):
    """
    Abstract base class for accumulators.

    Accumulators are used to accumulate values over multiple chunks.
    """

    def __init__(self, preprocess: Callable[[T], T] | None = maybe_hist) -> None:
        """
        Parameters
        ----------
        preprocess:
            Preprocessing function to be applied to pushed values prior to accumulation.
        """
        self._preprocess = preprocess

    def push(self, value: T) -> None:
        """
        Push a value to the accumulator.

        Parameters
        ----------
        value:
            Value to be pushed to the accumulator.
        """
        if self._preprocess is not None:
            value = self._preprocess(value)
        self._do_push(value)

    @abstractmethod
    def _do_push(self, value: T) -> None: ...

    @property
    @abstractmethod
    def value(self) -> T:
        """
        Get the accumulated value.

        Returns
        -------
        :
            Accumulated value.
        """


class EternalAccumulator(Accumulator[T]):
    """
    Simple accumulator that adds pushed values immediately.

    Does not support event data.
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._value: T | None = None

    @property
    def value(self) -> T:
        return self._value.copy()

    def _do_push(self, value: T) -> None:
        if self._value is None:
            self._value = value.copy()
        else:
            self._value += value


class RollingAccumulator(Accumulator[T]):
    """
    Accumulator that adds pushed values to a rolling window.

    Does not support event data.
    """

    def __init__(self, window: int = 10, **kwargs: Any) -> None:
        """
        Parameters
        ----------
        window:
            Size of the rolling window.
        """
        super().__init__(**kwargs)
        self._window = window
        self._values: list[T] = []

    @property
    def value(self) -> T:
        # Naive and potentially slow implementation if values and/or window are large!
        return sc.reduce(self._values).sum()

    def _do_push(self, value: T) -> None:
        self._values.append(value)
        if len(self._values) > self._window:
            self._values.pop(0)


class StreamProcessor:
    """
    Wrap a base workflow for streaming processing of chunks.

    Note that this class can not determine if the workflow is valid for streamed
    processing based on the input keys. In particular, it is the responsibility of the
    user to ensure that the workflow is "linear" with respect to the dynamic keys up to
    the accumulation keys.
    """

    def __init__(
        self,
        base_workflow: sciline.Pipeline,
        *,
        dynamic_keys: tuple[sciline.typing.Key, ...],
        target_keys: tuple[sciline.typing.Key, ...],
        accumulators: dict[sciline.typing.Key, Accumulator, Callable[..., Accumulator]]
        | tuple[sciline.typing.Key, ...],
    ) -> None:
        """
        Create a stream processor.

        Parameters
        ----------
        base_workflow:
            Workflow to be used for processing chunks.
        dynamic_keys:
            Keys that are expected to be updated with each chunk.
        target_keys:
            Keys to be computed and returned.
        accumulators:
            Keys at which to accumulate values and their accumulators. If a tuple is
            passed, :py:class:`EternalAccumulator` is used for all keys. Otherwise, a
            dict mapping keys to accumulator instances can be passed. If a dict value is
            a callable, base_workflow.bind_and_call(value) is used to make an instance.
        """
        workflow = sciline.Pipeline()
        for key in target_keys:
            workflow[key] = base_workflow[key]
        for key in dynamic_keys:
            workflow[key] = None  # hack to prune branches

        # Find and pre-compute static nodes as far down the graph as possible
        # See also https://github.com/scipp/sciline/issues/148.
        nodes = _find_descendants(workflow, dynamic_keys)
        parents = _find_parents(workflow, nodes) - nodes
        for key, value in base_workflow.compute(parents).items():
            workflow[key] = value

        self._process_chunk_workflow = workflow.copy()
        self._finalize_workflow = workflow.copy()
        self._accumulators = (
            accumulators
            if isinstance(accumulators, dict)
            else {key: EternalAccumulator() for key in accumulators}
        )
        # Depending on the target_keys, some accumulators can be unused and should not
        # be computed when adding a chunk.
        self._accumulators = {
            key: value
            for key, value in self._accumulators.items()
            if key in self._process_chunk_workflow.underlying_graph
        }
        # Create accumulators unless instances were passed. This allows for initializing
        # accumulators with arguments that depend on the workflow such as bin edges,
        # which would otherwise be hard to obtain.
        self._accumulators = {
            key: value
            if isinstance(value, Accumulator)
            else base_workflow.bind_and_call(value)
            for key, value in self._accumulators.items()
        }
        self._target_keys = target_keys

    def add_chunk(
        self, chunks: dict[sciline.typing.Key, Any]
    ) -> dict[sciline.typing.Key, Any]:
        for key, value in chunks.items():
            self._process_chunk_workflow[key] = value
            # There can be dynamic keys that do not "terminate" in any accumulator. In
            # that case, we need to make sure they can be and are used when computing
            # the target keys.
            self._finalize_workflow[key] = value
        to_accumulate = self._process_chunk_workflow.compute(self._accumulators)
        for key, processed in to_accumulate.items():
            self._accumulators[key].push(processed)
            self._finalize_workflow[key] = self._accumulators[key].value
        return self._finalize_workflow.compute(self._target_keys)


def _find_descendants(
    workflow: sciline.Pipeline, keys: tuple[sciline.typing.Key, ...]
) -> set[sciline.typing.Key]:
    graph = workflow.underlying_graph
    descendants = set()
    for key in keys:
        descendants |= nx.descendants(graph, key)
    return descendants | set(keys)


def _find_parents(
    workflow: sciline.Pipeline, keys: tuple[sciline.typing.Key, ...]
) -> set[sciline.typing.Key]:
    graph = workflow.underlying_graph
    parents = set()
    for key in keys:
        parents |= set(graph.predecessors(key))
    return parents



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import ipywidgets as ipw
import scipp as sc

UNITS_LIBRARY = {
    "wavelength": {"options": ("angstrom", "nm")},
    "Q": {"options": ("1/angstrom", "1/nm")},
    "Qx": {"options": ("1/angstrom", "1/nm")},
    "Qy": {"options": ("1/angstrom", "1/nm")},
    "tof": {"options": ("s", "ms", "us", "ns"), "selected": "us"},
    "dspacing": {"options": ("angstrom", "nm")},
    "energy_transfer": {"options": ("meV",)},
    "theta": {"options": ("rad", "deg")},
    "two_theta": {"options": ("rad", "deg")},
    "phi": {"options": ("rad", "deg")},
    "time": {"options": ("s", "ms", "us", "ns")},
    "temperature": {"options": ("K", "C", "F")},
}


class BinEdgesWidget(ipw.HBox, ipw.ValueWidget):
    def __init__(
        self,
        name: str,
        dim: str,
        start: float | None = None,
        stop: float | None = None,
        nbins: int = 1,
        unit: str | None = "undefined",
        log: bool = False,
    ):
        super().__init__()
        style = {
            "layout": {"width": "100px"},
            "style": {"description_width": "initial"},
        }
        units = UNITS_LIBRARY[dim] if unit == "undefined" else unit
        if isinstance(units, str):
            units = {"options": (units,)}
        if not isinstance(units, dict):
            units = {"options": units}
        self.fields = {
            "dim": ipw.Label(str(dim)),
            "unit": ipw.Dropdown(
                options=units["options"],
                value=units.get("selected", units["options"][0]),
                layout={"width": "initial"},
            ),
            "start": ipw.FloatText(description='start:', value=start, **style),
            "stop": ipw.FloatText(description='stop:', value=stop, **style),
            "nbins": ipw.BoundedIntText(
                description='nbins:', value=nbins, min=1, max=int(1e9), **style
            ),  # Note that a max value is required as it defaults to 100 without it.
            "spacing": ipw.Dropdown(
                options=['linear', 'log'],
                value='log' if log else 'linear',
                layout={"width": "initial"},
            ),
        }
        self.children = [ipw.Label(f"{name}:")] + list(self.fields.values())[1:]

    @property
    def value(self) -> sc.Variable:
        func = sc.geomspace if self.fields["spacing"].value == "log" else sc.linspace
        return func(
            dim=self.fields["dim"].value,
            start=self.fields["start"].value,
            stop=self.fields["stop"].value,
            num=self.fields["nbins"].value + 1,
            unit=self.fields["unit"].value,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import pytest
import scipp as sc
import scippnexus as snx
from scipp.testing import assert_identical

from ess.reduce import data
from ess.reduce.nexus import compute_component_position, workflow
from ess.reduce.nexus.types import (
    BackgroundRun,
    DetectorData,
    Filename,
    Monitor1,
    Monitor2,
    Monitor3,
    MonitorData,
    NeXusComponentLocationSpec,
    NeXusName,
    NeXusTransformation,
    SampleRun,
    TimeInterval,
)
from ess.reduce.nexus.workflow import (
    GenericNeXusWorkflow,
    LoadDetectorWorkflow,
    LoadMonitorWorkflow,
)


@pytest.fixture(params=[{}, {'aux': 1}])
def group_with_no_position(request) -> workflow.NeXusComponent[snx.NXsample, SampleRun]:
    return workflow.NeXusComponent[snx.NXsample, SampleRun](sc.DataGroup(request.param))


@pytest.fixture()
def depends_on() -> snx.TransformationChain:
    translation = snx.nxtransformations.Transform(
        name='/entry/instrument/comp1/transformations/trans1',
        transformation_type='translation',
        value=sc.scalar(1.0, unit='m'),
        vector=sc.vector(value=[1.0, 2.0, 3.0], unit=''),
        depends_on=snx.DependsOn(parent='', value='.'),
        offset=None,
    )
    return snx.TransformationChain(
        parent='/entry/instrument/comp1',
        value='transformations/trans1',
        transformations=sc.DataGroup({translation.name: translation}),
    )


@pytest.fixture()
def time_dependent_depends_on() -> snx.TransformationChain:
    """A chain of two transformations, the second one time-dependent."""
    trans1 = snx.nxtransformations.Transform(
        name='/entry/instrument/comp1/transformations/trans1',
        transformation_type='translation',
        value=sc.scalar(1.0, unit='m'),
        vector=sc.vector(value=[1.0, 0.0, 0.0], unit=''),
        depends_on=snx.DependsOn(
            parent='/entry/instrument/comp1/transformations', value='trans2'
        ),
    )
    trans2 = snx.nxtransformations.Transform(
        name='/entry/instrument/comp1/transformations/trans2',
        transformation_type='translation',
        value=sc.DataArray(
            sc.array(dims=['time'], values=[1.0, 2.0, 3.0], unit='m'),
            coords={'time': sc.array(dims=['time'], values=[0.0, 1.0, 2.0], unit='s')},
        ),
        vector=sc.vector(value=[0.0, 1.0, 0.0], unit=''),
        depends_on=snx.DependsOn(
            parent='/entry/instrument/comp1/transformations', value='.'
        ),
    )
    return snx.TransformationChain(
        parent='/entry/instrument/comp1',
        value='transformations/trans1',
        transformations=sc.DataGroup({trans1.name: trans1, trans2.name: trans2}),
    )


@pytest.fixture()
def transform(
    depends_on: snx.TransformationChain,
) -> NeXusTransformation[snx.NXdetector, SampleRun]:
    return NeXusTransformation.from_chain(depends_on)


def test_can_compute_position_of_group(depends_on: snx.TransformationChain) -> None:
    position = sc.vector([1.0, 2.0, 3.0], unit='m')
    group = workflow.NeXusComponent[snx.NXsource, SampleRun](
        sc.DataGroup(depends_on=depends_on)
    )
    chain = workflow.get_transformation_chain(group)
    trans = workflow.to_transformation(
        chain,
        interval=TimeInterval(slice(None, None)),
    )
    assert_identical(workflow.compute_position(trans), position)


def test_to_transform_with_positional_time_interval(
    time_dependent_depends_on: snx.TransformationChain,
) -> None:
    origin = sc.vector([0.0, 0.0, 0.0], unit='m')

    transform = workflow.to_transformation(
        time_dependent_depends_on,
        TimeInterval(slice(0, 1)),
    ).value
    assert sc.identical(transform * origin, sc.vector([1.0, 1.0, 0.0], unit='m'))

    transform = workflow.to_transformation(
        time_dependent_depends_on,
        TimeInterval(slice(1, 2)),
    ).value
    assert sc.identical(transform * origin, sc.vector([1.0, 2.0, 0.0], unit='m'))

    transform = workflow.to_transformation(
        time_dependent_depends_on,
        TimeInterval(slice(2, 3)),
    ).value
    assert sc.identical(transform * origin, sc.vector([1.0, 3.0, 0.0], unit='m'))


def test_to_transform_with_label_based_time_interval_single_point(
    time_dependent_depends_on: snx.TransformationChain,
) -> None:
    origin = sc.vector([0.0, 0.0, 0.0], unit='m')

    transform = workflow.to_transformation(
        time_dependent_depends_on,
        TimeInterval(slice(sc.scalar(0.1, unit='s'), sc.scalar(0.9, unit='s'))),
    ).value
    assert sc.identical(transform * origin, sc.vector([1.0, 1.0, 0.0], unit='m'))

    transform = workflow.to_transformation(
        time_dependent_depends_on,
        TimeInterval(slice(sc.scalar(1.1, unit='s'), sc.scalar(1.9, unit='s'))),
    ).value
    assert sc.identical(transform * origin, sc.vector([1.0, 2.0, 0.0], unit='m'))

    transform = workflow.to_transformation(
        time_dependent_depends_on,
        TimeInterval(slice(sc.scalar(2.1, unit='s'), sc.scalar(2.9, unit='s'))),
    ).value
    assert sc.identical(transform * origin, sc.vector([1.0, 3.0, 0.0], unit='m'))

    # No more new values after 2 seconds
    transform = workflow.to_transformation(
        time_dependent_depends_on,
        TimeInterval(slice(sc.scalar(1000.0, unit='s'), sc.scalar(2000.0, unit='s'))),
    ).value
    assert sc.identical(transform * origin, sc.vector([1.0, 3.0, 0.0], unit='m'))


def test_to_transform_raises_if_interval_does_not_yield_unique_value(
    time_dependent_depends_on: snx.TransformationChain,
) -> None:
    with pytest.raises(ValueError, match='Transform is time-dependent'):
        workflow.to_transformation(
            time_dependent_depends_on,
            TimeInterval(slice(sc.scalar(0.1, unit='s'), sc.scalar(1.9, unit='s'))),
        )


def test_given_no_sample_load_nexus_sample_returns_group_with_origin_depends_on() -> (
    None
):
    filespec = workflow.file_path_to_file_spec(
        data.loki_tutorial_sample_run_60250(), preopen=True
    )
    spec = workflow.unique_component_spec(filespec)
    assert spec.filename['/entry'][snx.NXsample] == {}
    sample = workflow.load_nexus_sample(spec)
    assert list(sample) == ['depends_on']
    chain = workflow.get_transformation_chain(sample)
    transformation = workflow.to_transformation(
        chain,
        interval=TimeInterval(slice(None, None)),
    )
    position = workflow.compute_position(transformation)
    assert_identical(position, sc.vector([0.0, 0.0, 0.0], unit='m'))


def test_get_transformation_chain_raises_exception_if_position_not_found(
    group_with_no_position,
) -> None:
    with pytest.raises(KeyError, match='depends_on'):
        workflow.get_transformation_chain(group_with_no_position)


@pytest.fixture()
def nexus_detector(
    depends_on: snx.TransformationChain,
) -> workflow.NeXusComponent[snx.NXdetector, SampleRun]:
    detector_number = sc.arange('detector_number', 6, unit=None)
    data = sc.DataArray(
        sc.empty_like(detector_number),
        coords={
            'detector_number': detector_number,
            'x_pixel_offset': sc.linspace('detector_number', 0, 1, num=6, unit='m'),
        },
    )
    return workflow.NeXusComponent[snx.NXdetector, SampleRun](
        sc.DataGroup(data=data, depends_on=depends_on, nexus_component_name='detector1')
    )


@pytest.fixture()
def source_position() -> sc.Variable:
    return sc.vector([0.0, 0.0, -10.0], unit='m')


def test_get_calibrated_detector_extracts_data_field_from_nexus_detector(
    nexus_detector, transform
) -> None:
    detector = workflow.get_calibrated_detector(
        nexus_detector, offset=workflow.no_offset, bank_sizes={}, transform=transform
    )
    assert_identical(detector, compute_component_position(nexus_detector)['data'])


def test_get_calibrated_detector_folds_detector_number_if_mapping_given(
    nexus_detector, transform
) -> None:
    sizes = {'xpixel': 2, 'ypixel': 3}
    bank_sizes = {'detector1': sizes}
    detector = workflow.get_calibrated_detector(
        nexus_detector,
        offset=workflow.no_offset,
        bank_sizes=bank_sizes,
        transform=transform,
    )
    assert detector.sizes == sizes


def test_get_calibrated_detector_works_if_nexus_component_name_is_missing(
    nexus_detector, transform
):
    del nexus_detector['nexus_component_name']
    detector = workflow.get_calibrated_detector(
        nexus_detector,
        offset=workflow.no_offset,
        bank_sizes={},
        transform=transform,
    )
    assert detector.sizes == nexus_detector['data'].sizes


def test_get_calibrated_detector_adds_offset_to_position(
    nexus_detector, transform
) -> None:
    offset = sc.vector([0.1, 0.2, 0.3], unit='m')
    detector = workflow.get_calibrated_detector(
        nexus_detector, offset=offset, bank_sizes={}, transform=transform
    )
    position = (
        compute_component_position(nexus_detector)['data'].coords['position'] + offset
    )
    assert detector.coords['position'].sizes == {'detector_number': 6}
    assert_identical(detector.coords['position'], position)


def test_get_calibrated_detector_position_dims_matches_data_dims(
    nexus_detector, transform
) -> None:
    nexus_detector2d = nexus_detector.fold('detector_number', sizes={'y': 2, 'x': 3})
    nexus_detector2d['data'].coords['x_pixel_offset'] = sc.linspace(
        'x', 0, 1, num=3, unit='m'
    )
    nexus_detector2d['data'].coords['y_pixel_offset'] = sc.linspace(
        'y', 0, 1, num=2, unit='m'
    )
    offset = sc.vector([0.1, 0.2, 0.3], unit='m')
    detector = workflow.get_calibrated_detector(
        nexus_detector2d, offset=offset, bank_sizes={}, transform=transform
    )
    assert detector.sizes == {'y': 2, 'x': 3}
    assert detector.coords['position'].sizes == {'y': 2, 'x': 3}


def test_get_calibrated_detector_position_unit_matches_offset_unit(
    nexus_detector, transform
) -> None:
    nexus_detector['data'].coords['x_pixel_offset'] = (
        nexus_detector['data'].coords['x_pixel_offset'].to(unit='mm')
    )
    offset = sc.vector([0.1, 0.2, 0.3], unit='m')
    detector = workflow.get_calibrated_detector(
        nexus_detector, offset=offset, bank_sizes={}, transform=transform
    )
    assert detector.coords['position'].unit == 'mm'


def test_get_calibrated_detector_forwards_coords(nexus_detector, transform) -> None:
    nexus_detector['data'].coords['abc'] = sc.scalar(1.2)
    detector = workflow.get_calibrated_detector(
        nexus_detector, offset=workflow.no_offset, bank_sizes={}, transform=transform
    )
    assert 'abc' in detector.coords


def test_get_calibrated_detector_forwards_masks(
    nexus_detector,
    transform,
) -> None:
    nexus_detector['data'].masks['mymask'] = sc.scalar(False)
    detector = workflow.get_calibrated_detector(
        nexus_detector, offset=workflow.no_offset, bank_sizes={}, transform=transform
    )
    assert 'mymask' in detector.masks


@pytest.fixture()
def calibrated_detector() -> workflow.CalibratedDetector[SampleRun]:
    detector_number = sc.arange('detector_number', 6, unit=None)
    return workflow.CalibratedDetector[SampleRun](
        sc.DataArray(
            sc.empty_like(detector_number),
            coords={
                'position': sc.vector([1.0, 2.0, 3.0], unit='m'),
                'detector_number': detector_number,
            },
        ).fold('detector_number', sizes={'xpixel': 2, 'ypixel': 3})
    )


@pytest.fixture()
def detector_event_data() -> workflow.NeXusData[snx.NXdetector, SampleRun]:
    content = sc.DataArray(
        sc.ones(dims=['event'], shape=[17], unit='counts'),
        coords={'event_id': sc.arange('event', 17, unit=None) % sc.index(6)},
    )
    weights = sc.bins(data=content, dim='event')
    return workflow.NeXusData[snx.NXdetector, SampleRun](
        sc.DataArray(
            weights,
            coords={
                'event_time_zero': sc.linspace(
                    dim=weights.dim, start=0, stop=1, num=weights.size, unit='s'
                )
            },
        )
    )


def test_assemble_detector_data_groups_events_by_detector_number(
    calibrated_detector, detector_event_data
) -> None:
    detector = workflow.assemble_detector_data(calibrated_detector, detector_event_data)
    assert detector.bins is not None
    assert_identical(
        detector.coords['detector_number'],
        calibrated_detector.coords['detector_number'],
    )
    # 17 events with arange%6 event_id, so 2 events in last bin
    assert_identical(
        detector.data.bins.size(),
        sc.array(dims=('xpixel', 'ypixel'), values=[[3, 3, 3], [3, 3, 2]], unit=None),
    )


def test_assemble_detector_data_does_not_add_event_id_coord(
    calibrated_detector, detector_event_data
) -> None:
    detector = workflow.assemble_detector_data(calibrated_detector, detector_event_data)
    assert 'event_id' not in detector.coords


def test_assemble_detector_data_adds_variances_to_weights(
    calibrated_detector, detector_event_data
) -> None:
    detector = workflow.assemble_detector_data(calibrated_detector, detector_event_data)
    assert detector_event_data.values[0].variances is None
    assert detector.values[0].variances is not None
    assert_identical(sc.variances(detector), sc.values(detector))


def test_assemble_detector_preserves_coords(calibrated_detector, detector_event_data):
    calibrated_detector.coords['abc'] = sc.scalar(1.2)
    detector = workflow.assemble_detector_data(calibrated_detector, detector_event_data)
    assert 'abc' in detector.coords


def test_assemble_detector_preserves_masks(calibrated_detector, detector_event_data):
    calibrated_detector.masks['mymask'] = sc.scalar(False)
    detector = workflow.assemble_detector_data(calibrated_detector, detector_event_data)
    assert 'mymask' in detector.masks


@pytest.fixture()
def nexus_monitor(
    depends_on: snx.TransformationChain,
) -> workflow.NeXusComponent[Monitor1, SampleRun]:
    data = sc.DataArray(sc.scalar(1.2), coords={'something': sc.scalar(13)})
    return workflow.NeXusComponent[Monitor1, SampleRun](
        sc.DataGroup(data=data, depends_on=depends_on)
    )


def test_get_calibrated_monitor_extracts_data_field_from_nexus_monitor(
    nexus_monitor,
) -> None:
    monitor = workflow.get_calibrated_monitor(
        nexus_monitor,
        offset=workflow.no_offset,
        source_position=sc.vector([0.0, 0.0, -10.0], unit='m'),
    )
    assert_identical(
        monitor.drop_coords(('position', 'source_position')),
        compute_component_position(nexus_monitor)['data'],
    )


def test_get_calibrated_monitor_subtracts_offset_from_position(
    nexus_monitor,
) -> None:
    offset = sc.vector([0.1, 0.2, 0.3], unit='m')
    monitor = workflow.get_calibrated_monitor(
        nexus_monitor,
        offset=offset,
        source_position=sc.vector([0.0, 0.0, -10.0], unit='m'),
    )
    assert_identical(monitor.coords['position'], sc.vector([1.1, 2.2, 3.3], unit='m'))


@pytest.fixture()
def calibrated_monitor() -> workflow.CalibratedMonitor[SampleRun, Monitor1]:
    return workflow.CalibratedMonitor[SampleRun, Monitor1](
        sc.DataArray(
            sc.scalar(0),
            coords={'position': sc.vector([1.0, 2.0, 3.0], unit='m')},
        )
    )


@pytest.fixture()
def monitor_event_data() -> workflow.NeXusData[Monitor1, SampleRun]:
    content = sc.DataArray(sc.ones(dims=['event'], shape=[17], unit='counts'))
    weights = sc.bins(data=content, dim='event')
    return workflow.NeXusData[Monitor1, SampleRun](
        sc.DataArray(
            weights,
            coords={
                'event_time_zero': sc.linspace(
                    dim=weights.dim, start=0, stop=1, num=weights.size, unit='s'
                )
            },
        )
    )


def test_assemble_monitor_data_adds_events_as_values_and_coords(
    calibrated_monitor, monitor_event_data
) -> None:
    monitor_data = workflow.assemble_monitor_data(
        calibrated_monitor, monitor_event_data
    )
    assert_identical(
        monitor_data.drop_coords(tuple(calibrated_monitor.coords)), monitor_event_data
    )


def test_assemble_monitor_data_adds_variances_to_weights(
    calibrated_monitor, monitor_event_data
) -> None:
    monitor_data = workflow.assemble_monitor_data(
        calibrated_monitor, monitor_event_data
    )
    assert_identical(
        sc.variances(monitor_data.drop_coords(tuple(calibrated_monitor.coords))),
        monitor_event_data,
    )


def test_assemble_monitor_preserves_coords(calibrated_monitor, monitor_event_data):
    calibrated_monitor.coords['abc'] = sc.scalar(1.2)
    monitor_data = workflow.assemble_monitor_data(
        calibrated_monitor, monitor_event_data
    )
    assert 'abc' in monitor_data.coords


def test_assemble_monitor_preserves_masks(calibrated_monitor, monitor_event_data):
    calibrated_monitor.masks['mymask'] = sc.scalar(False)
    monitor_data = workflow.assemble_monitor_data(
        calibrated_monitor, monitor_event_data
    )
    assert 'mymask' in monitor_data.masks


def test_load_monitor_workflow() -> None:
    wf = LoadMonitorWorkflow()
    wf[Filename[SampleRun]] = data.loki_tutorial_sample_run_60250()
    wf[NeXusName[Monitor1]] = 'monitor_1'
    da = wf.compute(MonitorData[SampleRun, Monitor1])
    assert 'position' in da.coords
    assert 'source_position' in da.coords
    assert da.bins is not None
    assert da.dims == ('event_time_zero',)


def test_load_detector_workflow() -> None:
    wf = LoadDetectorWorkflow()
    wf[Filename[SampleRun]] = data.loki_tutorial_sample_run_60250()
    wf[NeXusName[snx.NXdetector]] = 'larmor_detector'
    da = wf.compute(DetectorData[SampleRun])
    assert 'position' in da.coords
    assert 'sample_position' in da.coords
    assert 'source_position' in da.coords
    assert da.bins is not None
    assert da.dims == ('detector_number',)


def test_generic_nexus_workflow() -> None:
    wf = GenericNeXusWorkflow()
    wf[Filename[SampleRun]] = data.loki_tutorial_sample_run_60250()
    wf[NeXusName[Monitor1]] = 'monitor_1'
    wf[NeXusName[snx.NXdetector]] = 'larmor_detector'
    da = wf.compute(DetectorData[SampleRun])
    assert 'position' in da.coords
    assert 'sample_position' in da.coords
    assert 'source_position' in da.coords
    assert da.bins is not None
    assert da.dims == ('detector_number',)
    da = wf.compute(MonitorData[SampleRun, Monitor1])
    assert 'position' in da.coords
    assert 'source_position' in da.coords
    assert da.bins is not None
    assert da.dims == ('event_time_zero',)


def test_generic_nexus_workflow_raises_if_monitor_types_but_not_run_types_given() -> (
    None
):
    with pytest.raises(ValueError, match='run_types'):
        GenericNeXusWorkflow(monitor_types=[Monitor1])


def test_generic_nexus_workflow_includes_only_given_run_and_monitor_types() -> None:
    wf = GenericNeXusWorkflow(run_types=[SampleRun], monitor_types=[Monitor1, Monitor3])
    graph = wf.underlying_graph
    assert DetectorData[SampleRun] in graph
    assert DetectorData[BackgroundRun] not in graph
    assert MonitorData[SampleRun, Monitor1] in graph
    assert MonitorData[SampleRun, Monitor2] not in graph
    assert MonitorData[SampleRun, Monitor3] in graph
    assert MonitorData[BackgroundRun, Monitor1] not in graph
    assert MonitorData[BackgroundRun, Monitor2] not in graph
    assert MonitorData[BackgroundRun, Monitor3] not in graph
    # Many other keys are also removed, this is just an example
    assert NeXusComponentLocationSpec[Monitor1, SampleRun] in graph
    assert NeXusComponentLocationSpec[Monitor2, SampleRun] not in graph
    assert NeXusComponentLocationSpec[Monitor3, SampleRun] in graph
    assert NeXusComponentLocationSpec[snx.NXdetector, SampleRun] in graph
    assert NeXusComponentLocationSpec[snx.NXsample, SampleRun] in graph
    assert NeXusComponentLocationSpec[snx.NXsource, SampleRun] in graph
    assert NeXusComponentLocationSpec[Monitor1, BackgroundRun] not in graph
    assert NeXusComponentLocationSpec[Monitor2, BackgroundRun] not in graph
    assert NeXusComponentLocationSpec[Monitor3, BackgroundRun] not in graph
    assert NeXusComponentLocationSpec[snx.NXdetector, BackgroundRun] not in graph
    assert NeXusComponentLocationSpec[snx.NXsample, BackgroundRun] not in graph
    assert NeXusComponentLocationSpec[snx.NXsource, BackgroundRun] not in graph



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import scipp as sc
from ipywidgets import FloatText, GridBox, Label, Text, ValueWidget

from ..parameter import ParamWithBounds


class BoundsWidget(GridBox, ValueWidget):
    def __init__(self):
        super().__init__()

        self.fields = {
            'start': FloatText(description='start'),
            'end': FloatText(description='end'),
            'unit': Text(description='unit'),
        }
        self.children = [
            Label(value="Select bound:"),
            self.fields['unit'],
            self.fields['start'],
            self.fields['end'],
        ]

    @property
    def value(self):
        return (
            sc.scalar(self.fields['start'].value, unit=self.fields['unit']),
            sc.scalar(self.fields['end'].value, unit=self.fields['unit']),
        )

    @staticmethod
    def from_ess_parameter(_: ParamWithBounds) -> 'BoundsWidget':
        return BoundsWidget()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from typing import Any

from ipywidgets import HTML, HBox, Layout, RadioButtons, Widget

from ._config import default_style


class OptionalWidget(HBox):
    """Wrapper widget to handle optional widgets.

    When you retrieve the value of this widget,
    it will return the value of the wrapped widget.
    The parameter should be set as ``None`` or the actual value.
    """

    def __init__(self, wrapped: Widget, name: str = '', **kwargs) -> None:
        self.name = name
        self.wrapped = wrapped
        self._option_box = RadioButtons(
            description="",
            style=default_style,
            layout=Layout(width="auto", min_width="80px"),
            options={str(None): None, "": self.name},
        )
        self._option_box.value = None
        if hasattr(wrapped, "disabled"):
            # Disable the wrapped widget by default if possible
            # since the option box is set to None by default
            wrapped.disabled = True

        # Make the wrapped radio box horizontal
        self.add_class("widget-optional")
        _style_html = HTML(
            "<style>.widget-optional .widget-radio-box "
            "{flex-direction: row !important;} </style>",
            layout=Layout(display="none"),
        )

        def disable_wrapped(change) -> None:
            if change["new"] is None:
                if hasattr(wrapped, "disabled"):
                    wrapped.disabled = True
            else:
                if hasattr(wrapped, "disabled"):
                    wrapped.disabled = False

        self._option_box.observe(disable_wrapped, names="value")

        super().__init__([self._option_box, wrapped, _style_html], **kwargs)

    @property
    def value(self) -> Any:
        if self._option_box.value is None:
            self._option_box.value = None
            return None
        return self.wrapped.value

    @value.setter
    def value(self, value: Any) -> None:
        if value is None:
            self._option_box.value = None
        else:
            self._option_box.value = self.name
            self.wrapped.value = value



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from collections.abc import Callable, Generator
from contextlib import contextmanager
from typing import Any, NewType

import sciline as sl
import scipp as sc
from ipywidgets import FloatText, IntText

from ess.reduce.parameter import BinEdgesParameter, Parameter, parameter_registry
from ess.reduce.ui import WorkflowWidget, workflow_widget
from ess.reduce.widgets import OptionalWidget, SwitchWidget, create_parameter_widget
from ess.reduce.workflow import register_workflow, workflow_registry

SwitchableInt = NewType('SwitchableInt', int)
SwitchableFloat = NewType('SwitchableFloat', float)
OptionalInt = int | None
OptionalFloat = float | None


class IntParam(Parameter): ...


class FloatParam(Parameter): ...


parameter_registry[SwitchableInt] = IntParam('_', '_', 1, switchable=True)
parameter_registry[SwitchableFloat] = FloatParam('_', '_', 2.0, switchable=True)
parameter_registry[int] = IntParam('_', '_', 1)
parameter_registry[float] = FloatParam('_', '_', 2.0)
parameter_registry[OptionalInt] = IntParam('_', '_', 1, optional=True)
parameter_registry[OptionalFloat] = FloatParam('_', '_', 2.0, optional=True)


@create_parameter_widget.register(IntParam)
def _(param: IntParam) -> IntText:
    return IntText(value=param.default, description=param.name)


@create_parameter_widget.register(FloatParam)
def _(param: FloatParam) -> FloatText:
    return FloatText(value=param.default, description=param.name)


def _refresh_widget_parameter(
    *, widget: WorkflowWidget, output_selections: list[type]
) -> None:
    widget.output_selection_box.typical_outputs_widget.value = output_selections
    widget.parameter_box.parameter_refresh_button.click()


def _ready_widget(
    *,
    providers: list[Callable] | None = None,
    params: dict[type, Any] | None = None,
    output_selections: list[type],
    result_registry: dict[type, Any] | None = None,
) -> WorkflowWidget:
    widget = WorkflowWidget(
        sl.Pipeline(providers or [], params=params or {}),
        result_registry=result_registry,
    )
    _refresh_widget_parameter(widget=widget, output_selections=output_selections)
    return widget


def strict_provider(a: int, b: float) -> str:
    return f"{a} + {b}"


def provider_with_switch(a: SwitchableInt, b: SwitchableFloat) -> str:
    return f"{a} + {b}"


def provider_with_optional(a: OptionalInt, b: OptionalFloat) -> str:
    parts = [] if a is None else [str(a)]
    return ' + '.join([*parts] if b is None else [*parts, str(b)])


def _get_param_widget(widget: WorkflowWidget, param_type: type) -> Any:
    return widget.parameter_box._input_widgets[param_type].children[0]


def test_parameter_default_value_test() -> None:
    widget = _ready_widget(providers=[strict_provider], output_selections=[str])
    assert _get_param_widget(widget, int).value == 1
    assert _get_param_widget(widget, float).value == 2.0


def test_parameter_registry() -> None:
    assert isinstance(create_parameter_widget(IntParam('_a', '_a', 1)), IntText)
    assert isinstance(create_parameter_widget(FloatParam('_b', '_b', 2.0)), FloatText)


def test_run_not_allowed_when_parameter_not_refreshed_after_output_selected() -> None:
    widget = _ready_widget(providers=[strict_provider], output_selections=[str])
    # Clear the value of the output selection box
    widget.output_selection_box.typical_outputs_widget.value = []
    assert widget.result_box.run_button.disabled
    # Click the refresh button
    widget.parameter_box.parameter_refresh_button.click()
    assert not widget.result_box.run_button.disabled
    # Add a value to the parameter
    widget.output_selection_box.typical_outputs_widget.value = [str]
    assert widget.result_box.run_button.disabled
    # Click the refresh button again
    widget.parameter_box.parameter_refresh_button.click()
    assert not widget.result_box.run_button.disabled


def test_result_registry() -> None:
    registry = {}
    widget = _ready_widget(
        providers=[strict_provider], output_selections=[str], result_registry=registry
    )
    _get_param_widget(widget, int).value = 2
    _get_param_widget(widget, float).value = 0.1
    assert registry == {}
    widget.result_box.run_button.click()
    assert registry == {str: '2 + 0.1'}


def test_switchable_widget_dispatch() -> None:
    switchable_param = Parameter('a', 'a', 1, switchable=True)
    assert isinstance(create_parameter_widget(switchable_param), SwitchWidget)
    non_switchable_param = Parameter('b', 'b', 2, switchable=False)
    assert not isinstance(create_parameter_widget(non_switchable_param), SwitchWidget)


def test_switchable_parameter_switch_widget() -> None:
    widget = _ready_widget(providers=[provider_with_switch], output_selections=[str])

    int_widget = _get_param_widget(widget, SwitchableInt)
    float_widget = _get_param_widget(widget, SwitchableFloat)

    assert isinstance(int_widget, SwitchWidget)
    assert isinstance(float_widget, SwitchWidget)

    assert not float_widget.enabled
    assert not int_widget.enabled


def test_collect_values_from_disabled_switchable_widget() -> None:
    widget = _ready_widget(providers=[provider_with_switch], output_selections=[str])

    assert not _get_param_widget(widget, SwitchableFloat).enabled
    assert not _get_param_widget(widget, SwitchableInt).enabled
    assert widget.parameter_box.value == {}


def test_collect_values_from_enabled_switchable_widget() -> None:
    widget = _ready_widget(providers=[provider_with_switch], output_selections=[str])

    float_widget = _get_param_widget(widget, SwitchableFloat)
    float_widget.enabled = True
    float_widget.value = 0.2

    assert widget.parameter_box.value == {SwitchableFloat: 0.2}


def test_switchable_optional_parameter_switchable_first() -> None:
    dummy_param = Parameter('a', 'a', 1, switchable=True, optional=True)
    dummy_widget = create_parameter_widget(dummy_param)
    assert isinstance(dummy_widget, SwitchWidget)
    assert isinstance(dummy_widget.wrapped, OptionalWidget)


def test_optional_widget_dispatch() -> None:
    optional_param = Parameter('a', 'a', 1, optional=True)
    assert isinstance(create_parameter_widget(optional_param), OptionalWidget)
    non_optional_param = Parameter('b', 'b', 2, optional=False)
    assert not isinstance(create_parameter_widget(non_optional_param), OptionalWidget)


def test_optional_parameter_optional_widget() -> None:
    widget = _ready_widget(providers=[provider_with_optional], output_selections=[str])

    int_widget = _get_param_widget(widget, OptionalInt)
    float_widget = _get_param_widget(widget, OptionalFloat)

    assert isinstance(int_widget, OptionalWidget)
    assert isinstance(float_widget, OptionalWidget)

    assert float_widget.value is None
    assert int_widget.value is None


def test_collect_values_from_optional_widget() -> None:
    widget = _ready_widget(providers=[provider_with_optional], output_selections=[str])

    float_widget = _get_param_widget(widget, OptionalFloat)
    float_widget.value = 0.2

    assert widget.parameter_box.value == {OptionalFloat: 0.2, OptionalInt: None}


def test_collect_values_from_optional_widget_compute_result() -> None:
    result_registry = {}
    widget = _ready_widget(
        providers=[provider_with_optional],
        output_selections=[str],
        result_registry=result_registry,
    )

    float_widget = _get_param_widget(widget, OptionalFloat)
    float_widget.value = 0.2
    widget.result_box.run_button.click()

    assert result_registry == {str: '0.2'}

    int_widget = _get_param_widget(widget, OptionalInt)
    int_widget.value = 2
    widget.result_box.run_button.click()

    assert result_registry == {str: '2 + 0.2'}


def dummy_workflow_constructor() -> sl.Pipeline:
    return sl.Pipeline([strict_provider])


@contextmanager
def temporary_workflow_registry(
    *constructors: Callable[[], sl.Pipeline],
) -> Generator[None, None, None]:
    existance_flags = {
        constructor: constructor in workflow_registry for constructor in constructors
    }
    for constructor in constructors:
        register_workflow(constructor)
    yield
    for constructor, flag in existance_flags.items():
        if not flag:
            workflow_registry.discard(constructor)


def test_register_workflow() -> None:
    with temporary_workflow_registry(dummy_workflow_constructor):
        assert dummy_workflow_constructor in workflow_registry

    assert dummy_workflow_constructor not in workflow_registry


def _get_selection_widget(widget):
    return widget.children[0].children[0]


def test_workflow_registry_applied_to_selector() -> None:
    expected_constructor_pair = (
        'dummy_workflow_constructor',
        dummy_workflow_constructor,
    )
    with temporary_workflow_registry(dummy_workflow_constructor):
        selection_widget = _get_selection_widget(workflow_widget())
        assert expected_constructor_pair in selection_widget.options

    selection_widget = _get_selection_widget(workflow_widget())
    assert expected_constructor_pair not in selection_widget.options


def dummy_second_workflow_constructor() -> sl.Pipeline:
    return sl.Pipeline([provider_with_switch])


def test_workflow_selection() -> None:
    # Prepare
    with temporary_workflow_registry(
        dummy_workflow_constructor, dummy_second_workflow_constructor
    ):
        widget = workflow_widget()
        selection_widget = _get_selection_widget(widget)
        # Before selection
        assert len(widget.children[1].children) == 0
        # Select first workflow
        selection_widget.value = dummy_workflow_constructor
        assert len(widget.children[1].children) == 1
        # Test created WorkflowWidget
        first_widget = widget.children[1].children[0]
        assert isinstance(first_widget, WorkflowWidget)
        assert first_widget.output_selection_box.typical_outputs_widget.options == (
            ('str', str),
        )
        _refresh_widget_parameter(widget=first_widget, output_selections=[str])
        assert first_widget.parameter_box._input_widgets.keys() == {int, float}
        # Select second workflow
        selection_widget.value = dummy_second_workflow_constructor
        second_widget = widget.children[1].children[0]
        _refresh_widget_parameter(widget=second_widget, output_selections=[str])
        assert second_widget.parameter_box._input_widgets.keys() == {
            SwitchableInt,
            SwitchableFloat,
        }


WavelengthBins = NewType('WavelengthBins', sc.Variable)
WavelengthBinsWithUnit = NewType('WavelengthBinsWithUnit', sc.Variable)
QBins = NewType('QBins', sc.Variable)
TemperatureBinsWithUnits = NewType('TemperatureBinsWithUnits', sc.Variable)


parameter_registry[WavelengthBins] = BinEdgesParameter(WavelengthBins, dim='wavelength')
parameter_registry[WavelengthBinsWithUnit] = BinEdgesParameter(
    WavelengthBinsWithUnit, dim='wavelength', unit='m'
)
parameter_registry[QBins] = BinEdgesParameter(
    QBins, dim='Q', start=0.01, stop=0.6, nbins=150
)
parameter_registry[TemperatureBinsWithUnits] = BinEdgesParameter(
    TemperatureBinsWithUnits, dim='temperature', unit=('Kelvin', 'Celsius', 'Degrees')
)


def wavelength_bins_print_provider(bins: WavelengthBins) -> str:
    return str(bins)


def wavelength_bins_with_unit_print_provider(bins: WavelengthBinsWithUnit) -> str:
    return str(bins)


def q_bins_print_provider(bins: QBins) -> str:
    return str(bins)


def temperature_bins_print_provider(bins: TemperatureBinsWithUnits) -> str:
    return str(bins)


def test_bin_edges_widget_simple() -> None:
    widget = _ready_widget(
        providers=[wavelength_bins_print_provider], output_selections=[str]
    )
    param_widget = _get_param_widget(widget, WavelengthBins)
    assert sc.identical(
        param_widget.value,
        sc.linspace(dim='wavelength', start=0.0, stop=0.0, num=2, unit='angstrom'),
    )
    assert set(param_widget.fields['unit'].options) == {'angstrom', 'nm'}
    assert param_widget.fields['nbins'].value == 1
    assert param_widget.fields['spacing'].value == 'linear'
    # Modify the values
    param_widget.fields['start'].value = 1.0
    param_widget.fields['stop'].value = 200.0
    param_widget.fields['nbins'].value = 10
    assert sc.identical(
        param_widget.value,
        sc.linspace(dim='wavelength', start=1.0, stop=200.0, num=11, unit='angstrom'),
    )


def test_bin_edges_widget_override_unit() -> None:
    widget = _ready_widget(
        providers=[wavelength_bins_with_unit_print_provider], output_selections=[str]
    )
    param_widget = _get_param_widget(widget, WavelengthBinsWithUnit)
    assert sc.identical(
        param_widget.value,
        sc.linspace(dim='wavelength', start=0.0, stop=0.0, num=2, unit='m'),
    )
    assert set(param_widget.fields['unit'].options) == {'m'}
    assert param_widget.fields['nbins'].value == 1


def test_bin_edges_widget_override_unit_multiple() -> None:
    widget = _ready_widget(
        providers=[temperature_bins_print_provider], output_selections=[str]
    )
    param_widget = _get_param_widget(widget, TemperatureBinsWithUnits)
    assert sc.identical(
        param_widget.value,
        sc.linspace(dim='temperature', start=0.0, stop=0.0, num=2, unit='Kelvin'),
    )
    assert set(param_widget.fields['unit'].options) == {'Kelvin', 'Celsius', 'Degrees'}


def test_bin_edges_widget_log_spacing() -> None:
    widget = _ready_widget(
        providers=[wavelength_bins_print_provider], output_selections=[str]
    )
    param_widget = _get_param_widget(widget, WavelengthBins)
    param_widget.fields['spacing'].value = 'log'
    param_widget.fields['start'].value = 1.0e3
    param_widget.fields['stop'].value = 1.0e5
    param_widget.fields['nbins'].value = 7
    assert sc.identical(
        param_widget.value,
        sc.geomspace(dim='wavelength', start=1.0e3, stop=1.0e5, num=8, unit='angstrom'),
    )
    assert param_widget.fields['spacing'].value == 'log'


def test_bin_edges_widget_with_default_values() -> None:
    widget = _ready_widget(providers=[q_bins_print_provider], output_selections=[str])
    param_widget = _get_param_widget(widget, QBins)
    assert sc.identical(
        param_widget.value,
        sc.linspace(dim='Q', start=0.01, stop=0.6, num=151, unit='1/angstrom'),
    )
    assert set(param_widget.fields['unit'].options) == {'1/angstrom', '1/nm'}
    assert param_widget.fields['start'].value == 0.01
    assert param_widget.fields['stop'].value == 0.6
    assert param_widget.fields['nbins'].value == 150
    assert param_widget.fields['spacing'].value == 'linear'



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
"""Tools for creating live data reduction workflows for Beamlime."""

from pathlib import Path
from typing import NewType, TypeVar

import sciline
import scipp as sc
import scippnexus as snx

from ess.reduce import streaming
from ess.reduce.nexus import types as nt
from ess.reduce.nexus.json_nexus import JSONGroup

JSONEventData = NewType('JSONEventData', dict[str, JSONGroup])


def _load_json_event_data(name: str, nxevent_data: JSONEventData) -> sc.DataArray:
    return snx.Group(nxevent_data[name], definitions=snx.base_definitions())[()]


def load_json_event_data_for_sample_run(
    name: nt.NeXusName[nt.Component], nxevent_data: JSONEventData
) -> nt.NeXusData[nt.Component, nt.SampleRun]:
    return nt.NeXusData[nt.Component, nt.SampleRun](
        _load_json_event_data(name, nxevent_data)
    )


def load_json_event_data_for_sample_transmission_run(
    name: nt.NeXusName[nt.Component], nxevent_data: JSONEventData
) -> nt.NeXusData[nt.Component, nt.TransmissionRun[nt.SampleRun]]:
    return nt.NeXusData[nt.Component, nt.TransmissionRun[nt.SampleRun]](
        _load_json_event_data(name, nxevent_data)
    )


T = TypeVar('T', bound='LiveWorkflow')


class LiveWorkflow:
    """A workflow class that fulfills Beamlime's LiveWorkflow protocol."""

    def __init__(
        self,
        *,
        streamed: streaming.StreamProcessor,
        outputs: dict[str, sciline.typing.Key],
    ) -> None:
        self._streamed = streamed
        self._outputs = outputs

    @classmethod
    def from_workflow(
        cls: type[T],
        *,
        workflow: sciline.Pipeline,
        accumulators: dict[sciline.typing.Key, streaming.Accumulator],
        outputs: dict[str, sciline.typing.Key],
        run_type: type[nt.RunType],
        nexus_filename: Path,
    ) -> T:
        """
        Create a live workflow from a base workflow and other parameters.

        Parameters
        ----------
        workflow:
            Base workflow to use for live data reduction.
        accumulators:
            Accumulators forwarded to the stream processor.
        outputs:
            Mapping from output names to keys in the workflow. The keys correspond to
            workflow results that will be computed.
        run_type:
            Type of the run to process. This defines which run is the dynamic run being
            processed. The NeXus template file will be set as the filename for this run.
        nexus_filename:
            Path to the NeXus file to process.

        Returns
        -------
        :
            Live workflow object.
        """

        workflow = workflow.copy()
        if run_type is nt.SampleRun:
            workflow.insert(load_json_event_data_for_sample_run)
        elif run_type is nt.TransmissionRun[nt.SampleRun]:
            workflow.insert(load_json_event_data_for_sample_transmission_run)
        else:
            raise NotImplementedError(f"Run type {run_type} not supported yet.")
        workflow[nt.Filename[run_type]] = nexus_filename
        streamed = streaming.StreamProcessor(
            base_workflow=workflow,
            dynamic_keys=(JSONEventData,),
            target_keys=outputs.values(),
            accumulators=accumulators,
        )
        return cls(streamed=streamed, outputs=outputs)

    def __call__(
        self, nxevent_data: dict[str, JSONGroup], nxlog: dict[str, JSONGroup]
    ) -> dict[str, sc.DataArray]:
        """
        Implements the __call__ method required by the LiveWorkflow protocol.

        Parameters
        ----------
        nxevent_data:
            NeXus event data.
        nxlog:
            NeXus log data. WARNING: This is currently not used.

        Returns
        -------
        :
            Dictionary of computed and plottable results.
        """
        # Beamlime passes full path, but the workflow only needs the name of the monitor
        # or detector group.
        nxevent_data = {
            key.lstrip('/').split('/')[2]: value for key, value in nxevent_data.items()
        }
        results = self._streamed.add_chunk({JSONEventData: nxevent_data})
        return {name: results[key] for name, key in self._outputs.items()}
